'''
Created on Jul 23, 2019

@author: chengzi
'''

# imports
from pathlib import Path
from PIL import Image, ImageFile
ImageFile.LOAD_TRUNCATED_IMAGES = True

from torch.utils.data import Dataset, DataLoader, Subset
from torch.utils.data import sampler

import torchvision.transforms as T


class CocoDataset(Dataset):
    def __init__(self, path_dataset, transform=None):
        p_iter = Path(path_dataset).iterdir()
        self.transform = transform
        self.images_list = [name for name in p_iter]
    
    def __len__(self):
        '''provides the size of the dataset'''
        return len(self.images_list)
    
    def __getitem__(self, idx):
        '''
        supporting integer indexing in range from 0 to len(self) exclusive
        '''
        im = Image.open(self.images_list[idx]).convert("L")  # to grayscale
        if self.transform:
            im = self.transform(im)
        return im

# function to denomalize normalized image
def denormalize(im, mean, std):
    '''
    im: pytorch tensor view as image
    '''
    assert len(im.size()) == 3
    if im.shape[0] == 1: # grayscale
#         im = im.reshape(im.shape[0], im.shape[1])
        im = im * std[0] + mean[0]
    else: # rgb
        for ch in range(3):
            im[:,:,ch] = im[:,:,ch] * std[ch] + mean[ch]
    return im

# Test if CocoDataset is OK
def coco_dataset(input_width_height):
    path_dset_train = '/media/chengzi/Win-D/DataSet-20180112/COCO2014/coco/images/train2014'
    path_dset_test = '/media/chengzi/Win-D/DataSet-20180112/COCO2014/coco/images/val2014'
    
    path_dset_train = '/media/chengzi/FT-data/BIBM2019/cocodemo/cdotrain2014'
    path_dset_test = '/media/chengzi/FT-data/BIBM2019/cocodemo/cdotest2014'
            
    train_set = CocoDataset(path_dset_train,
                             transform=T.Compose(
                                [
                                    T.RandomResizedCrop(input_width_height),
#                                     T.RandomGrayscale(),
                                    T.RandomHorizontalFlip(),
                                    T.RandomVerticalFlip(),
                                    T.ToTensor(),
                                    T.Normalize((0.5,), (0.5,))
                                ])
                            )
    test_set = CocoDataset(path_dset_test,
                           transform=T.Compose(
                                [
                                    T.RandomResizedCrop(input_width_height),
                                    T.ToTensor(),
                                    T.Normalize((0.5,), (0.5,))
                                ])
                          )
    

    im = train_set[5]
    print(im.shape)
    im = denormalize(im, (0.5,), (0.5,))
#     _im = im.reshape((im.shape[1], im.shape[2]))
#     plt.imshow(_im, cmap='gray')
#     plt.show()
    print(len(train_set), len(test_set))
    
    TOTAL_SAMPLES = len(train_set)
    NUM_TRAIN = TOTAL_SAMPLES // 5 * 4
    BATCH_SIZE = 32
    
    loader_train = DataLoader(train_set, batch_size=BATCH_SIZE, sampler=sampler.SubsetRandomSampler(range(NUM_TRAIN)), drop_last=False)
    loader_subtrain = DataLoader(Subset(train_set, [i for i in range(3)]), batch_size=BATCH_SIZE)  # used for checking avg_psnr
    loader_val = DataLoader(train_set, batch_size=BATCH_SIZE, sampler=sampler.SubsetRandomSampler(range(NUM_TRAIN, TOTAL_SAMPLES)))
    loader_test = DataLoader(test_set, batch_size=1,num_workers=1)
    
   
    return loader_train, loader_subtrain, loader_val, loader_test
    
if __name__ == '__main__':  
    coco_dataset()

        