'''
https://github.com/y0umu/ResCSNet

@author: chengzi
'''

import torch.nn as nn
import numpy as np
import torch
from ResCSNet.LoadCOCO import coco_dataset

input_width_height = 120
        
# original linear encoder
class EncoderLinear(nn.Module):
    def _init_weights(self, m):
#         print(m)
        if type(m) == nn.Conv2d:
            nn.init.kaiming_normal_(m.weight.data)
            if m.bias is not None:
                nn.init.constant_(m.bias.data, 0)
    
    def __init__(self, N, M):
        super().__init__()
        self.linear = nn.Linear(N, M)
        self.linear.apply(self._init_weights)
    
    def forward(self, x):
        x_vec = x.view(x.shape[0], 1, 1, -1)
        out = self.linear(x_vec)
        return out
    
  
class TheUpsample(nn.Module):
    '''
    Wrapper for torch.nn.functional.interpolate
    '''
    def __init__(self, size):
        super().__init__()
        self.size = size
    
    def forward(self, x):
        return nn.functional.interpolate(x, size=self.size, mode='nearest')

class ResCsNet(nn.Module):
    def __init__(self, N, M):
#     def __init__(self, encoder_out_ch, encoder_ksize, encoder_stride, encoder_padding):
        super().__init__()

        WIDTH_HALF = int(input_width_height/4*2)
        WIDTH_THREE_QUARTERS = int(input_width_height/4*3)
        # encoder
        self.encoder= EncoderLinear(N, M)
        
        # upsample
        self.initial_width = int(np.ceil(np.sqrt(M))) # should be identical to WIDTH_INITIAL
        self.upsample0 = TheUpsample((1, int(self.initial_width**2)) )
#         upsample1_width = int(input_width_height/4)
#         self.upsample1 = TheUpsample((upsample1_width, upsample1_width))
        upsample2_width = WIDTH_HALF
        self.upsample2 = TheUpsample((upsample2_width, upsample2_width))
        upsample3_width = WIDTH_THREE_QUARTERS
        self.upsample3 = TheUpsample((upsample3_width, upsample3_width))
        # the last upsample should up sample to the original size
        self.upsample4 = TheUpsample((input_width_height, input_width_height))
        
        # conv_scale
        self.conv_scale1 = nn.Conv2d(96, 1, kernel_size=1, stride=1, padding=0)
        self.conv_scale2 = nn.Conv2d(96, 1, kernel_size=1, stride=1, padding=0)
        self.conv_scale3 = nn.Conv2d(96, 1, kernel_size=1, stride=1, padding=0)
        self.conv_scale4 = nn.Conv2d(96, 1, kernel_size=1, stride=1, padding=0)
        
        # units
        # unit 1
        self.unit1 = nn.Sequential(
            nn.Conv2d(1, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
            nn.LeakyReLU()

        )

        # unit2
        self.unit2a_branch1 = nn.Sequential(
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
        )
        
        self.unit2a_branch2 = nn.Sequential(
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
            nn.LeakyReLU(),
            
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
            nn.LeakyReLU(),
            
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96)
        )
        
        self.unit2b_branch2 = nn.Sequential(
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
            nn.LeakyReLU(),
            
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
            nn.LeakyReLU(),
            
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96)
        )        
        
        # unit 3
        # branch1 is identical pass
        self.unit3_branch2 = nn.Sequential(
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
            nn.LeakyReLU(),
            
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
            nn.LeakyReLU(),
            
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96)
        )
        
        # unit 4
        self.unit4_branch1 = nn.Sequential(
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
        )
        
        self.unit4_branch2 = nn.Sequential(
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
            nn.LeakyReLU(),
            
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
            nn.LeakyReLU(),
            
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96)
        )
        
        # unit 5
        # branch 1 is identical pass
        self.unit5_branch2 = nn.Sequential(
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
            nn.LeakyReLU(),
            
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96),
            nn.LeakyReLU(),
            
            nn.Conv2d(96, 96, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(96)
        )
    
    def forward(self, x):
        x = self.encoder(x)
        x_vec = self.upsample0(x)
        x = x_vec.view(x.shape[0], 1, self.initial_width, self.initial_width)  # 23x23, 32x32, 45x45, 55x55
        
        # unit 1
        x = self.unit1(x)
#         x = self.upsample1(x)
        
        # unit 2
        x2a_1 = self.unit2a_branch1(x)
        x2a_2 = self.unit2a_branch2(x)
        x = x2a_1 + x2a_2
        x2b_1 = x.clone()
        x2b_2 = self.unit2b_branch2(x)
        x = x2b_1 + x2b_2
        
        x = self.upsample2(x)  # remove this layer for r=0.3
        # unit 3
        x3_1 = x.clone()
        x3_2 = self.unit3_branch2(x)
        x = x3_1 + x3_2
        
        x = self.upsample3(x)
        # unit 4
        x4_1 = self.unit4_branch1(x)
        x4_2 = self.unit4_branch2(x)
        x = x4_1 + x4_2
        
        x = self.upsample4(x)
        # unit5
        x5_1 = x.clone()
        x5_2 = self.unit5_branch2(x)
        x = x5_1 + x5_2
        im4 = self.conv_scale4(x)
        
        return im4


def ResCsNet_init_weights(m):
        '''
        used to init weights. This is just a clone of the method in the class above. Provided for convenience
        '''
#         print(m)
        if type(m) == nn.Conv2d:
            nn.init.kaiming_normal_(m.weight.data)
            if m.bias is not None:
                nn.init.constant_(m.bias.data, 0)
        if type(m) == nn.Linear:
            nn.init.kaiming_normal_(m.weight.data)
            nn.init.constant_(m.bias.data, 0)
        if type(m) == nn.ConvTranspose2d:
            nn.init.kaiming_normal_(m.weight.data)
            if m.bias is not None:
                nn.init.constant_(m.bias.data, 0)

############################################
# make instance and test the model
def test_encoder():
    loader_train, loader_subtrain, loader_val, loader_test = coco_dataset()
    
    input_width_height = 100
    N = input_width_height*input_width_height
    M = int(N * 0.35)
    encoder = EncoderLinear(N, M)
    im = torch.randn(64,1,input_width_height,input_width_height)
    
    for t, original_im in enumerate(loader_train):
        im = original_im
        y = encoder(im)
        print(y.size())
        
    y = encoder(im)
    print(y.size())
    
    
if __name__ == '__main__':  
    test_encoder()
