'''
Created on Jul 23, 2019

@author: chengzi
'''

import torch
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
import torch.nn.functional as F
from ResCSNet.resCSnet_model import ResCsNet
from ResCSNet.LoadCOCO import coco_dataset

device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
print('using device:', device)

    
# Visualize the recovered and original image
def denormalize(im):
    if im.shape[2] == 1: # grayscale
        mean, std = (0.5,), (0.5,)
        im = im.reshape(im.shape[0], im.shape[1])
        im = im * std[0] + mean[0]
    else: # rgb
        mean, std = (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
        for ch in range(3):
            im[:,:,ch] = im[:,:,ch] * std[ch] + mean[ch]
    return im

# visualize recovered image
def vis_show(loader_test, N, M, modelname, idx=0):
   
    model = ResCsNet(N, M)
    model.load_state_dict(torch.load(modelname))
    model.eval()
    model.to(device)
    print('trained model loaded')
    
    for _,img in enumerate(loader_test):
        with torch.no_grad():
            img = img.to(device=torch.device("cuda"))
            restored_im = model(img)
        print(restored_im.shape)
    
        _im = restored_im[idx].cpu().numpy()
        _im = _im.transpose(1, 2, 0).astype(np.float)
        _ori = img[idx].cpu().numpy()
        _ori = _ori.transpose(1, 2, 0).astype(np.float)

        mse = np.mean((_im - _ori) ** 2 )
        psnr = 10 * np.log10(1/mse)
        print("mse is", mse, "psnr is", psnr, "dB")

        _im = np.clip(_im*255.0, 0, 255).astype(np.uint8)  # must be converted into np.uint8 then show correct images
        _im = Image.fromarray(_im.reshape(_im.shape[0],_im.shape[1]), 'L')
        _ori = np.clip(_ori*255.0, 0, 255).astype(np.uint8)  # must be converted into np.uint8 then show correct images
        _ori = Image.fromarray(_ori.reshape(_ori.shape[1],_ori.shape[0]), 'L')
        
        plt.subplot(1,2,1)
        plt.imshow(_ori, cmap="gray"); plt.title('original')
        plt.grid(False)
        plt.subplot(1,2,2)
        plt.imshow(_im, cmap="gray"); plt.title('restored')
        plt.grid(False)
        plt.pause(0.6)
    
def ResCSNet_eval(loader_test,N, M,modelname):

    model = ResCsNet(N, M)
    model.load_state_dict(torch.load(modelname))
    model.eval()
    model.to(device)
    print('trained model loaded')       

    avg_psnr = 0
    _psnrs = []
    with torch.no_grad():
        for _,batch in enumerate(loader_test):
            original = batch.to(device)
            recovered = model(original)
                
            recovered = denormalize(recovered)
            original = denormalize(original)
#             diff = recovered - original
#             rmse = np.sqrt( np.mean(diff ** 2 ) )
#             psnr = 20 * np.log10(1/rmse)
            mse = F.mse_loss(recovered, original)
            psnr = 10 * np.log10(1 / mse.item())
            _psnrs.append(psnr)
            avg_psnr += psnr
    print("===> Avg. PSNR: {:.4f} dB".format(avg_psnr / len(loader_test)))
    
    psnrs = np.array(_psnrs)
    plt.hist(psnrs, bins=15); plt.xlabel('PSNR/dB'); plt.ylabel('Number of samples')
    plt.show()

if __name__ == '__main__':
    
    SAMPLING_RATE = 0.10
    input_width_height = 120
    N = input_width_height*input_width_height
    M = int(N * SAMPLING_RATE)
    _, _, _, loader_test = coco_dataset(input_width_height)

    modelname = '/home/chengzi/Desktop/workspace20170624/CSMRI/ResCSNet/ResCsNet-colab_0.10.pth'
    
#     ResCSNet_eval(loader_test,N,M,modelname)
    vis_show(loader_test,N,M,modelname)
    